/* /////////////////////////////////////////////////////////////////////////////
 * File:        glob.c
 *
 * Purpose:     Definition of the glob() API functions for the Win32 platform.
 *
 * Created      13th November 2002
 * Updated:     3rd November 2003
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/
 *                          http://www.synesis.com.au/software/
 *
 *              email:      software@synesis.com.au
 *                          software@synesis-group.com
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#define _SYNSOFT_VER_C_GLOB_MAJOR       1
#define _SYNSOFT_VER_C_GLOB_MINOR       2
#define _SYNSOFT_VER_C_GLOB_REVISION    1
#define _SYNSOFT_VER_C_GLOB_EDIT        23

#include <windows.h>
#include <errno.h>

#include "glob.h"

#ifndef NUM_ELEMENTS
# define NUM_ELEMENTS(x)        (sizeof(x) / sizeof(0[x]))
#endif /* !NUM_ELEMENTS */

/* /////////////////////////////////////////////////////////////////////////////
 * Helper functions
 */

static char const *strrpbrk(char const *string, char const *strCharSet)
{
    char        *part   =   NULL;
    char const  *pch;

    for(pch = strCharSet; *pch; ++pch)
    {
        char    *p  =   strrchr(string, *pch);

        if(NULL != p)
        {
            if(NULL == part)
            {
                part = p;
            }
            else
            {
                if(part < p)
                {
                    part = p;
                }
            }
        }
    }

    return part;
}

/* /////////////////////////////////////////////////////////////////////////////
 * API functions
 */

/* It gives you back the matched contents of your pattern, so for Win32, the
 * directories must be included
 */

int glob(   char const  *pattern
        ,   int         flags
        ,   const int (*errfunc)(char const *, int)
        ,   glob_t      *pglob)
{
    int         result;
    char        szRelative[1 + _MAX_PATH];
    char const  *file_part  =   strrpbrk(pattern, "\\/");

    if(NULL != file_part)
    {
        ++file_part;

        lstrcpyA(szRelative, pattern);
        szRelative[file_part - pattern] = '\0';
    }
    else
    {
        szRelative[0] = '\0';
    }

#if 0
    if(!GetFullPathNameA(pattern, NUM_ELEMENTS(szRelative), szRelative, &file_part))
    {
        result = GLOB_NOMATCH;
    }
    else
#endif /* 0 */
    {
        WIN32_FIND_DATAA    find_data;
        HANDLE              hFind = FindFirstFileA(pattern, &find_data);

        pglob->gl_pathc = 0;
        pglob->gl_pathv = NULL;

        if(hFind == INVALID_HANDLE_VALUE)
        {
            result = GLOB_NOMATCH;
        }
        else
        {
            int     cbCurr      =   0;
            size_t  cbAlloc     =   0;
            char    *buffer     =   NULL;
            int     cMatches    =   0;

            result = 0;

            do
            {
                int     cch;
                size_t  new_cbAlloc;

                if(find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
                {
                    if(NULL != file_part)
                    {
                        /* Patter must begin with '.' to match either dots directory */
                        if( *file_part != '.' &&
                            0 == lstrcmpA(".", find_data.cFileName) ||
                            0 == lstrcmpA("..", find_data.cFileName))
                        {
                            continue;
                        }
                    }

                    if(flags & GLOB_MARK)
                    {
                        lstrcatA(find_data.cFileName, "/");
                    }
                }
                else
                {
                    if(flags & GLOB_ONLYDIR)
                    {
                        // Skip all further actions, and get the next entry
                        continue;
                    }
                }

                cch =   lstrlenA(find_data.cFileName);
                cch +=  file_part - pattern;

                new_cbAlloc = cbCurr + cch + 1;
                if(new_cbAlloc > cbAlloc)
                {
                    char    *new_buffer;

                    new_cbAlloc = (new_cbAlloc + 31) & ~(31);

                    new_buffer  = (char*)realloc(buffer, new_cbAlloc);

                    if(new_buffer == NULL)
                    {
                        result = GLOB_NOSPACE;
                        free(buffer);
                        buffer = NULL;
                        break;
                    }

                    buffer = new_buffer;
                    cbAlloc = new_cbAlloc;
                }

                lstrcpynA(buffer + cbCurr, szRelative, 1 + (file_part - pattern));
                lstrcatA(buffer + cbCurr, find_data.cFileName);
                cbCurr += cch + 1;

                ++cMatches;
            }
            while(FindNextFile(hFind, &find_data));

            FindClose(hFind);

            if(result == 0)
            {
                char    *new_buffer =   (char*)realloc(buffer, cbAlloc + (1 + cMatches) * sizeof(char*));

                if(new_buffer == NULL)
                {
                    result = GLOB_NOSPACE;
                    free(buffer);
                    buffer = NULL;
                }
                else
                {
                    char    **pp    = (char**)new_buffer;
                    char    **begin =   pp;
                    char    **end   =   begin + cMatches;
                    char    *next_str;

                    buffer = new_buffer;

                    memmove(new_buffer + (1 + cMatches) * sizeof(char*), new_buffer, cbAlloc);

                    if(flags & GLOB_NOSORT)
                    {
                        /* The way we need in order to test the removal of dots in the findfile_sequence. */
                        *end = NULL;
                        for(begin = pp, next_str = buffer + (1 + cMatches) * sizeof(char*); begin != end; --end)
                        {
                            *(end - 1) = next_str;

                            // Find the next string.
                            next_str += 1 + lstrlenA(next_str);
                        }
                    }
                    else
                    {
                        /* The normal way. */
                        for(begin = pp, next_str = buffer + (1 + cMatches) * sizeof(char*); begin != end; ++begin)
                        {
                            *begin = next_str;

                            // Find the next string.
                            next_str += 1 + lstrlenA(next_str);
                        }
                        *begin = NULL;
                    }

                    pglob->gl_pathc =   cMatches;
                    pglob->gl_matchc=   cMatches;
                    pglob->gl_offs  =   0;
                    pglob->gl_flags =   0;
                    pglob->gl_pathv =   pp;
                }
            }
        }
    }

    return result;
}

void globfree(glob_t *pglob)
{
    if(pglob != NULL)
    {
        free(pglob->gl_pathv);
        pglob->gl_pathv = NULL;
    }
}

/* ////////////////////////////////////////////////////////////////////////// */
